{&Use32+}
program show_irq_link_conflicts;

(* option to emulate a different machine using pirt.dat and pci.log *)
{- $Define pci_Debug}

uses
  pirt,
  {$IfDef VirtualPascal}
  VpUtils,
  {$Else VirtualPascal}
  TpUtils,
  {$EndIf VirtualPascal}
  Crt,
  RedirCon,
  pci_hw,
  pci_ca;


var
  irq_used,
  irq_used_bridge       :array[$a..$d] of boolean;
                      (*       Low(pirt_type.slotentrytable)..High(pirt_type.slotentrytable) *)
  bridgeonly            :array[2                            ..$7ff] of boolean;
  sl                    :word;
  link,pin              :byte;
  found                 :boolean;
  irq_line_bitmap_and   :smallword;
  irq_line_bitmap_or    :smallword;

{$I via_bios.inc}
{$I classes.pas}

function r_vendor(const w:word):boolean;
  begin
    r_vendor:=(pirt_ptr^.compatible_vendor=w) or (infotbl_W($00)=w);
  end;

function r_device(const w:word):boolean;
  begin
    r_device:=(pirt_ptr^.compatible_device=w) or (infotbl_W($02)=w);
  end;

function linkinfo(link:byte):string;
  var
    b                   :byte;
    via_index,
    via_shift           :byte;
  begin
    with pirt_ptr^ do
      begin
        infotbl_bus:=router_bus;
        infotbl_deviceid:=router_devfunc shr 3;
        infotbl_func:=router_devfunc and 7;
        reset_infotbl_cache;
        linkinfo:=' (no link info, ask for compatible:$'+Int2Hex(compatible_vendor,4)+':$'+Int2Hex(compatible_device,4)+')';

        (* Intel, 82371FB PIIX PCI to ISA Bridge,.. *)
        if  (link in [1..4,$60..$63])
        and r_vendor($8086)
        and (   r_device($7110)  (* compatible: 7000! *)
             or r_device($122e)
             or r_device($2480)  (* 82801CA LPC Interface (ICH3-S B1 step) *)
             or r_device($248C)  (* 82801CAM (ICH3-M) LPC Interface *)
             or r_device($1234)  (* 82371MX 430MX Mobile Chipset MPIIX + EIDE + I/O -- found in compatible to $248C *)
             or r_device($244C)  (* 82801BAM LPC Interface (ICH2-M B2 step) *)
                               )
         then
          begin
            (* 29056201.pdf Seite 59:
               Bit7     Interrupt routing enable 0=enable 1=disable
                  6..4  reserved, read:0
                  3..0  IRQ 3/4/5/6/7/9/10/11/12/14/15 oder reserved *)
            if link in [1..4] then
              b:=infotbl($60+link-1)
            else
              b:=infotbl(link);
            if Odd(b shr 7) then
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end (* Intel *)

        (* Intel canterwood *)
        (* Intel 845GL *)
        else
        if (link in [$60..$63,$68..$6b])
        and r_vendor($8086)
        and (   r_device($24d0)  (* 82801EB/ER (ICH5/ICH5R) LPC Interface Bridge *)
             or r_device($7000)  (* 845: compatible to ...  7000: 82371SB PIIX3 ISA Bridge *)
             or r_device($2410)  (* 82801AA 8xx Chipset LPC Interface Bridge *)
             or r_device($24CC)) (* 82801DBM (ICH4-M) LPC Interface Bridge *)
         then
          begin
            b:=infotbl(link);
            if (b<1) or (b>=$80) then (* guessed.. *)
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end

        (* VIA, VT82C596/A/B "Mobile South" PCI to ISA Bridge,.. *)
        else
        if (   ((link in [1..9]) and tranlate_via_link(link,via_index,via_shift))
            or (link in [$55,$56,$57]))
        and r_vendor($1106)
        and (   r_device($0586)
             or r_device($0596)
             or r_device($0686)
             or r_device($0696))
         then
          begin
(* IL(1a/b1/via):
Format of AMD-645 Peripheral Bus Controller, function 0 (PCI-ISA bridge) data:
54h    BYTE    PCI IRQ Edge/Level selection (see #01024)
55h    BYTE    PnP Routing for external MIRQ0/1 (see #01025)
56h    BYTE    PnP Routing for PCI INTB/INTA (see #01027)
57h    BYTE    PnP Routing for PCI INTD/INTC (see #01028)
58h    BYTE    PnP Routing for external MIRQ2 (see #01029)
59h    BYTE    MIRQ pin configuration (see #01030)

Bitfields for AMD-645 PCI IRQ Edge/Level Select register:
Bit(s)  Description     (Table 01024)
 7-4    reserved
 3      PIRQA# is edge-sensitive rather than level-sensitive
 2      PIRQB# is edge-sensitive
 1      PIRQC# is edge-sensitive
 0      PIRQD# is edge-sensitive
SeeAlso: #01011,#01023,#01025

Bitfields for AMD-645 PnP IRQ Routing 1 register:
Bit(s)  Description     (Table 01025)
 7-4    routing for MIRQ1 (see #01026)
 3-0    routing for MIRQ0 (see #01026)
SeeAlso: #01011,#01024

(Table 01026)
Values for AMD-645 IRQ routing:
 0000   disabled
 0001   IRQ1
 0010   reserved
 0011   IRQ3
 ...
 0111   IRQ7
 1000   reserved
 1001   IRQ9
 ...
 1100   IRQ12
 1101   reserved
 1110   IRQ14
 1111   IRQ15
SeeAlso: #01025,#01027,#01028,#01029

Bitfields for AMD-645 PnP IRQ Routing 2 register:
Bit(s)  Description     (Table 01027)
 7-4    routing for PIRQB# (see #01026)
 3-0    routing for PIRQA# (see #01026)
SeeAlso: #01025,#01028,#01029

Bitfields for AMD-645 PnP IRQ Routing 3 register:
Bit(s)  Description     (Table 01028)
 7-4    routing for PIRQD# (see #01026)
 3-0    routing for PIRQC# (see #01026)
SeeAlso: #01025,#01027,#01029

Bitfields for AMD-645 PnP IRQ Routing 4 register:
Bit(s)  Description     (Table 01029)
 7-4    reserved
 3-0    routing for MIRQ2# (see #01026)
SeeAlso: #01025,#01027,#01028

Bitfields for AMD-645 MIRQ Pin Configuration register:
Bit(s)  Description     (Table 01030)
 7-3    reserved (0)
 2      select MASTER# instead of MIRQ2
 1      select KEYLOCK instead of MIRQ1
 0      select APICCS# instead of MIRQ0
SeeAlso: #01011,#01029                                  *)
(*
            case link of
                1:b:=infotbl($56) shr 4;
                2:b:=infotbl($56) shr 0;
                3:b:=infotbl($57) shr 4;
                4:b:=infotbl($57) shr 0; {?}
                5:b:=infotbl($55) shr 4;
                6:b:=infotbl($55) shr 0; {?}
              $55:b:=infotbl($55) shr 4;
              $56:b:=infotbl($56) shr 0;{?? shr 0/4}
              $57:b:=infotbl($57) shr 4;
            end;*)

            case link of
              $55:
                begin
                  via_index:=$55;
                  via_shift:=4;
                end;
              $56:
                begin
                  via_index:=$56;
                  via_shift:=0;
                end;
              $57:
                begin
                  via_index:=$57;
                  via_shift:=4;
                end;
            end;

            b:=(infotbl(via_index) shr via_shift) and $0f;

            if b=0 then
              linkinfo:=' (IRQ disabled)'
             else
              linkinfo:=' IRQ '+Int2Str(b);

            {
            if link in [Low(via_translate)..High(via_translate)] then
              link:=via_translate[link];
            case link of
              1:b:=(infotbl($55) shr 4) and $0f;
              2:b:=(infotbl($56) shr 0) and $0f;
              3:b:=(infotbl($56) shr 4) and $0f;
           // 4:b:=??
              5:b:=(infotbl($44) shr 0) and $0f;
              6:b:=(infotbl($44) shr 4) and $0f;
              7:b:=(infotbl($45) shr 0) and $0f;
              8:b:=(infotbl($45) shr 4) and $0f;
            //9:b:=(infotbl($45)?
            else
                b:=0;
            end;}
          end (* VIA *)


        (* SiS635,SiS730 *)
        else
        if  (link in [$41,$42,$43,$44])
        and r_vendor($1039)
        and (   r_device($0008))
        (* more?
        D       0008    SiS PCI to ISA Bridge (LPC Bridge)
        D       0018    SiS950 PCI to ISA Bridge (LPC Bridge)
        D       0963    SiS963 PCI to ISA Bridge (LPC Bridge)
        D       5582    SiS5582 PCI to ISA Bridge *)
         then
          begin
            b:=infotbl(link);
            if Odd(b shr 7) then
              linkinfo:=' IRQ '+Int2Str(b-$80)
            else
              linkinfo:=' (IRQ disabled)'
          end
        (* SiS496/497 *)
        else
        if  (link in [1..4,$c0..$c3])
        and r_vendor($1039)
        and r_device($0496)
         then
          begin
            if link<$c0 then
              b:=infotbl($c0+link-1)
            else
              b:=infotbl(link);
            if Odd(b shr 7) then
              linkinfo:=' IRQ '+Int2Str(b-$80)
            else
              linkinfo:=' (IRQ disabled)'
          end


        (* NVidia Corp - guessed *)
        else
        if  (link in [1..{8}4])
        and r_vendor($10de)
         then
          begin
            case link of
              1:b:=infotbl($7c) shr 0;
              2:b:=infotbl($7c) shr 4;
              3:b:=infotbl($7d) shr 0;
              4:b:=infotbl($7d) shr 4;
            end;
            b:=b and $f;
            if b=0 then
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end


        (* guessed *)
        (* D 1533 ALI M1533 Aladdin IV/V ISA South Bridge *)
        else
        if  (link in [1..4])
        and r_vendor($10b9)
        and r_device($1533)
         then
          begin
            b:=infotbl($48+(link-1) div 2);
            if Odd(link-1) then
              b:=b shr 4;
            b:=b and $f;

            case b of
            (*
            LINE 0 1 2 3 4 5 6 7 8 9 a b c d e f
            ALI  0 0 0 2 4 5 7 6 0 1 3 9 b 0 d f *)
              0:b:= 0;
              1:b:= 9;
              2:b:= 3;
              3:b:=$a;
              4:b:= 4;
              5:b:= 5;
              6:b:= 7;
              7:b:= 6;
              8:b:=-0;
              9:b:=$b;
             $a:b:=-0;
             $b:b:=$c;
             $c:b:=-0;
             $d:b:=$e;
             $e:b:=-0;
             $f:b:=$f;
            end;
            if b=0 then
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end

        (* ALi MAGiK 1 - guessed *)
        else
        if  (link in [1..8{9?}])
        and r_vendor($10b9)
         then
          begin
            b:=infotbl($48+(link-1) div 2);
            if Odd(link-1) then
              b:=b shr 4;
            b:=b and $f;

            case b of
            (*
            LINE 0 1 2 3 4 5 6 7 8 9 a b c d e f
            ALI  0 8 0 2 4 5 7 6 0 1 3 9 b 0 d f *)
              0:b:= 0;
              1:b:= 9;
              2:b:= 3;
              3:b:=$a;
              4:b:= 4;
              5:b:= 5;
              6:b:= 7;
              7:b:= 6;
              8:b:= 1;
              9:b:=$b;
             $a:b:=-0;
             $b:b:=$c;
             $c:b:=-0;
             $d:b:=$e;
             $e:b:=-0;
             $f:b:=$f;
            end;
            if b=0 then
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end

        (* Cyrix 5530 *)
        else
        if  (link in [1..4])
        and r_vendor($1078)
        and (r_device($0000{unsure}) or r_device($0002) or r_device($0100))
        (* D 0000 Cx5520 ISA Bridge Rev.0 *)
        (* D 0002 Cx5520 ISA Bridge Rev.1 *)
        (* D 0100 5530 Kahula/Geode Legacy ISA Bridge *)
         then
          begin
            b:=infotbl_L($5c) shr (4*link-1);
            if b=0 then
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end

        (* AMD *)
        else
        if  (link in [1..4])
        and r_vendor($1022)
        and (   r_device($7410)  (* AMD-766 PCI to ISA/LPC Bridge *)
             or r_device($7400)  (* AMD-755 (Cobra) PCI to ISA Bridge -geraten fr AMD-75X: k7ai2aa1.bin *)
             or r_device($7408)) (* AMD-756 (Viper) PCI to ISA Bridge -geraten fr AMD-75X: k7ai2aa1.bin *)
(* ? D  7440    AMD-768 PCI to ISA/LPC Bridge *)
         then
          begin
            b:=infotbl_l($56) shr (4*link-1);
            if b=0 then
              linkinfo:=' (IRQ disabled)'
            else
              linkinfo:=' IRQ '+Int2Str(b);
          end

        (* nothing found *)
        else
          begin
            (* keep 'no link info' *)
          end;

      end;
  end;

procedure display_classcode;
  var
    pcl                 :word;
  begin
    if (infotbl($0b)>=Low(pci_class_names)) and (infotbl($0b)<=High(pci_class_names)) then
      Write(pci_class_names[infotbl($0b)],':')
    else
      Write('Class=',infotbl($0b),'?:');
    pcl:=Low(pci_class_array);
    repeat
      if pcl>High(pci_class_array) then
        begin
          (* map 'Storage:Subclass=1 programming interface=138'
          (* to  'Storage:IDE(programming interface=138)' *)
          pcl:=Low(pci_class_array);
          repeat
            if pcl>High(pci_class_array) then
              begin
                Write('Subclass=',infotbl($0a),' programming interface=',infotbl($09));
                Break;
              end;
            with pci_class_array[pcl] do
              if (class_=infotbl($0b)) and (subclass=infotbl($0a)) then
                begin
                  Write(cname);
                  (* programming interface is special for storage/ide:
                     it has busmaster/native mode.. flags              *)
                  if (infotbl($0b)<>1(*Stoarge*)) or (infotbl($0a)<>1(*IDE*)) then
                    Write(' (programming interface=',infotbl($09),')');
                  Break;
                end;
            Inc(pcl);
          until false;
          Break;
        end;

      with pci_class_array[pcl] do
        if (class_=infotbl($0b)) and (subclass=infotbl($0a)) and (progif=infotbl($09)) then
          begin
            Write(cname);
            Break;
          end;
      Inc(pcl);
    until false;
  end; (* display_classcode *)

procedure dipalay_intline;
  begin
    if WhereX<40 then GotoXY(40,WhereY);
    if infotbl($3c) in [1..254{15?}] then
      Write(' (',infotbl($3c),')')
    else (* 0/255 *)
      Write(' (disabled)');
  end;

procedure search_devices(var used_a,used_b,used_c,used_d:boolean;const bus_min,bus_max:byte);
  var
    org_bus,
    org_deviceid,
    org_func            :byte;
  begin
    org_bus     :=infotbl_bus;
    org_deviceid:=infotbl_deviceid;
    org_func    :=infotbl_func;

    infotbl_bus:=bus_min;
    for infotbl_deviceid:=0 to 31 do
      for infotbl_func:=0 to 7 do
        begin
          reset_infotbl_cache;

          if infotbl_W($00)<>$ffff then
            begin

              case infotbl($3d) of
                1:used_a:=true;
                2:used_b:=true;
                3:used_c:=true;
                4:used_d:=true;
              end;


              if is_bridge_with_pass_pci_irq then
                if (infotbl_bus<infotbl($19)) (* secondary bus ok? *)
                and (infotbl($19)<=infotbl($1a)) (* subordinate bus number ok? *)
                and (infotbl($1a)<=bus_max) then
                  search_devices(used_a,used_b,used_c,used_d,infotbl($19),infotbl($1a));

            end;

          if not is_multifunction then
            Break;
        end;

    infotbl_bus     :=org_bus;
    infotbl_deviceid:=org_deviceid;
    infotbl_func    :=org_func;
    reset_infotbl_cache;
  end;

procedure display_devices(const pin,bus_min,bus_max:byte);
  var
    org_bus,
    org_deviceid,
    org_func            :byte;
  begin
    org_bus     :=infotbl_bus;
    org_deviceid:=infotbl_deviceid;
    org_func    :=infotbl_func;

    infotbl_bus:=bus_min;
    for infotbl_deviceid:=0 to 31 do
      for infotbl_func:=0 to 7 do
        begin
          reset_infotbl_cache;

          if infotbl_W($00)<>$ffff then
            begin

              if infotbl($3d)-1+$a=pin then
                begin
                  WriteLn('* Bus=',infotbl_bus:3,' Device=',infotbl_deviceid shr 3:2,'             INT',Int2Hex(pin,1),'#');
                  Write('  Function=',infotbl_func,'  ');
                  display_classcode;
                  dipalay_intline;
                  WriteLn;
                end;


              if is_bridge_with_pass_pci_irq then
                if (infotbl_bus<infotbl($19)) (* secondary bus ok? *)
                and (infotbl($19)<=infotbl($1a)) (* subordinate bus number ok? *)
                and (infotbl($1a)<=bus_max) then
                  display_devices(pin,infotbl($19),infotbl($1a));

            end;

          if not is_multifunction then
            Break;
        end;

    infotbl_bus     :=org_bus;
    infotbl_deviceid:=org_deviceid;
    infotbl_func    :=org_func;
    reset_infotbl_cache;
  end;


begin

  {$IfDef OS2}
  if ioredirected then
    begin
      Assign(Output,'');
      Rewrite(Output);
    end
  else
    begin
      ClrScr;
      install_pager;
    end;
  {$EndIf}

  if (ParamStr(1)='?') or (ParamStr(1)='-?') or (ParamStr(1)='/?')
  or (ParamCount>1) then
    begin
      WriteLn('Show PCI Routing Table from link view');
      WriteLn('Usage: SHOW_LNK [BIOS-source]');
      WriteLn('  if no BIOS source file is specified, it is taken from the running BIOS,');
      WriteLn('  and SHOW_LNK als interprets it by looking for matching PCI devices.');

      Halt(99);
    end;

  open_pci_access_driver;

{$IfDef pci_Debug2}
  get_pir('M:\pirt.dat');
{$Else}
  get_pir(ParamStr(1));
{$EndIf}
  if not Assigned(pirt_ptr) then
    begin
      WriteLn('table not found.');
      {$IfDef OS2}
      if ParamStr(1)='' then
        WriteLn('please try to run me in a DOS session.');
      {$EndIf OS2}
      Halt(99);
    end;

  with pirt_ptr^ do
    begin

      if ParamStr(1)='' then
        begin
          pci_present_test;
          if failed then
            begin
              WriteLn('PCI not present.');
              Halt(99);
            end;


          for sl:=Low(slotentrytable) to table_size div SizeOf(slot_entry_type)-1 do
            with slotentrytable[sl] do
              begin
                FillChar(irq_used,SizeOf(irq_used),false);
                FillChar(irq_used_bridge,SizeOf(irq_used_bridge),false);
                infotbl_bus:=bus;
                infotbl_deviceid:=device shr 3;
                for infotbl_func:=0 to 7 do
                  begin
                    reset_infotbl_cache;

                    if infotbl_W(0)<>$ffff then
                      begin

                        if infotbl($3d) in [1..4] then
                          irq_used[infotbl($3d)-1+$a]:=true;

                        if is_bridge_with_pass_pci_irq then
                          if (infotbl_bus<infotbl($19)) (* secondary bus ok? *)
                          and (infotbl($19)<=infotbl($1a)) (* subordinate bus number ok? *)
                           then
                            search_devices(irq_used_bridge[$a],
                                           irq_used_bridge[$b],
                                           irq_used_bridge[$c],
                                           irq_used_bridge[$d],infotbl($19),infotbl($1a));

                      end;

                    if not is_multifunction then
                      Break;

                  end;

                (* delete slot entries with no present devices *)
                for pin:=$a to $d do
                  if not (irq_used[pin] or irq_used_bridge[pin]) then
                    begin
                      link_bitmap[pin].link_value:=0;
                      link_bitmap[pin].irq_bitmap:=0;
                    end;

                (* mark bridge only pin use *)
                bridgeonly[sl]:=true;
                for pin:=$a to $d do
                  if irq_used[pin] then
                    bridgeonly[sl]:=false;

              end;
        end;

      for link:=1 to 255 do
        begin
          found:=false;
          irq_line_bitmap_and:=$ffff; (* all  *)
          irq_line_bitmap_or :=$0000; (* none *)
          for sl:=Low(slotentrytable) to table_size div SizeOf(slot_entry_type)-1 do
            with slotentrytable[sl] do
              for pin:=Low(link_bitmap) to High(link_bitmap) do
                with link_bitmap[pin] do
                  if link=link_value then
                    begin
                      if not found then
                        begin
                          Write('Link value $',Int2Hex(link,2),':');
                          if ParamStr(1)='' then
                            Write(linkinfo(link));
                          WriteLn;
                          found:=true;
                        end;
                      Write('+ Bus=',bus:3,' Device=',device shr 3:2);
                      if slot_number=0 then
                        Write(' (mainboard) ')
                      else
                        Write(' (slot=',slot_number:2,')   ');

                      if bridgeonly[sl] then
                        begin
                          if ParamStr(1)='' then
                            begin
                              (* assume: bridge is on func 0 *)
                              infotbl_bus:=bus;
                              infotbl_deviceid:=device shr 3;
                              infotbl_func:=0;
                              reset_infotbl_cache;
                              display_classcode;
                            end
                          else
                            Write('bridge')
                        end
                      else
                        Write('INT',Int2Hex(pin,1),'#');
                      WriteLn;
                      irq_line_bitmap_and:=irq_line_bitmap_and and irq_bitmap;
                      irq_line_bitmap_or :=irq_line_bitmap_or  or  irq_bitmap;
                      if ParamStr(1)='' then
                        begin
                          infotbl_bus:=bus;
                          infotbl_deviceid:=device shr 3;
                          for infotbl_func:=0 to 7 do
                            begin
                              reset_infotbl_cache;
                              (* no device? *)
                              if infotbl_W(0)<>$ffff then
                                begin

                                  if infotbl($3d)=pin-$a+1 then
                                    begin
                                      Write('  Function=',infotbl_func,'  ');
                                      display_classcode;
                                      dipalay_intline;
                                      WriteLn;
                                    end;

                                  if is_bridge_with_pass_pci_irq then
                                    if (infotbl_bus<infotbl($19)) (* secondary bus ok? *)
                                    and (infotbl($19)<=infotbl($1a)) (* subordinate bus number ok? *)
                                     then
                                      display_devices(pin,infotbl($19),infotbl($1a));
                                end;

                              if not is_multifunction then
                                Break;
                            end; (* info_func *)
                        end; (* ParamStr(1)='' *)

                    end;
          if found then
            begin
              WriteLn('> possible connection: ',irq_bitmap_str(irq_line_bitmap_and,irq_line_bitmap_or));
              WriteLn;
            end;
        end;

    end;

  Dispose(pirt_ptr);
  close_pci_access_driver;
end.

